package org.jboss.cache.eviction;

import org.jboss.cache.CacheSPI;
import org.jboss.cache.Fqn;
import org.jboss.cache.config.EvictionPolicyConfig;

/**
 * @author Ben Feb 13, 2004
 */
public class DummyEvictionPolicy extends BaseEvictionPolicy
{
   public void evict(Fqn fqn) throws Exception
   {
      // no-op
//      throw new RuntimeException("Testing only");
   }

   public int getWakeupIntervalSeconds()
   {
      return 0;  //To change body of implemented methods use File | Settings | File Templates.
   }

   public void setCache(CacheSPI cache)
   {
      // no op
   }

   public EvictionAlgorithm getEvictionAlgorithm()
   {
      // no op
      return null;
   }

   public Class<? extends EvictionPolicyConfig> getEvictionConfigurationClass()
   {
      return null;
   }


}
