package org.jboss.cache.transaction.isolationlevels;

import org.jboss.cache.lock.IsolationLevel;

/**
 * @author <a href="mailto:manik@jboss.org">Manik Surtani</a>
 * @since 2.0.0
 */
public class ReadUncommittedTest extends IsolationLevelTestBase
{
   public ReadUncommittedTest()
   {
      isolationLevel = IsolationLevel.READ_UNCOMMITTED;
   }
}
