// $Id: UpgradeException.java 3152 2006-12-11 21:14:35Z genman $

/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 * Created on Jan 18, 2003
 */
package org.jboss.cache.lock;


/**
 * Used when a read-lock cannot be upgraded to a write-lock
 *
 * @author Bela Ban
 * @version $Revision: 3152 $
 */
public class UpgradeException extends LockingException
{

   private static final long serialVersionUID = -5683212651728276028L;

   /**
    * Constructor for UpgradeException.
    *
    * @param msg
    */
   public UpgradeException(String msg)
   {
      super(msg);
   }

   public UpgradeException(String msg, Throwable cause)
   {
      super(msg, cause);
   }

}
