package org.jboss.cache.api;

import org.testng.annotations.Test;

/**
 * An optimistic version of {@link org.jboss.cache.api.NodeAPITest}
 */
@Test(groups = "functional")
public class NodeAPIOptimisticTest extends NodeAPITest
{
   public NodeAPIOptimisticTest()
   {
      optimistic = true;
   }
}
