/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.jmx;


/**
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
public class OptimisticNotificationTest extends NotificationTest
{
   public OptimisticNotificationTest()
   {
      optimistic = true;
   }
}
