package org.jboss.cache.lock;

import java.util.Date;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;

class NullLock implements Lock, Condition {

   public void lock()
   {
   }

   public void lockInterruptibly() throws InterruptedException
   {
   }

   public Condition newCondition()
   {
      return this;
   }

   public boolean tryLock() 
   {
      return true;
   }

   public boolean tryLock(long arg0, TimeUnit arg1) throws InterruptedException
   {
      return true;
   }

   public void unlock()
   {
   }

   public void await() throws InterruptedException
   {
   }

   public boolean await(long arg0, TimeUnit arg1) throws InterruptedException
   {
      return true;
   }

   public long awaitNanos(long arg0) throws InterruptedException
   {
      return arg0;
   }

   public void awaitUninterruptibly()
   {
   }

   public boolean awaitUntil(Date arg0) throws InterruptedException
   {
      return true;
   }

   public void signal()
   {
   }

   public void signalAll()
   {
   }
}
