package org.jboss.cache.loader.s3;

import org.jboss.cache.CacheException;

/**
 * Basic exception class.
 */
public class S3Exception extends CacheException
{

   private static final long serialVersionUID = -5961236335942313217L;

   /**
    * Constructs a new S3Exception.
    */
   public S3Exception()
   {
   }

   /**
    * Constructs a new S3Exception.
    */
   public S3Exception(String arg0)
   {
      super(arg0);
   }

   /**
    * Constructs a new S3Exception.
    */
   public S3Exception(Throwable arg0)
   {
      super(arg0);
   }

   /**
    * Constructs a new S3Exception.
    */
   public S3Exception(String arg0, Throwable arg1)
   {
      super(arg0, arg1);
   }

}
