package org.jboss.cache.api.nodevalidity;

import org.testng.annotations.Test;

/**
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani</a>
 * @since 2.1.0
 */
@Test(groups = {"functional"})
public class ReplicatedOptNodeValidityTest extends ReplicatedPessNodeValidityTest
{
   public ReplicatedOptNodeValidityTest()
   {
      optimistic = true;
   }
}