package org.jboss.cache.interceptors;

/**
 * MBean to the {@link InvocationContextInterceptor}
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani</a>
 */
public interface InvocationContextInterceptorMBean extends InterceptorMBean
{
}
