/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.buddyreplication;

import org.testng.annotations.Test;

/**
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
@Test(groups = {"functional", "jgroups"})
public class BuddyReplicationWithOptimisticLockingTest extends BuddyReplicationFailoverTest
{
   public BuddyReplicationWithOptimisticLockingTest()
   {
      optimisticLocks = true;
   }

   @Override
   public void testDataGravitationKillOwner() throws Exception
   {
      super.testDataGravitationKillOwner();
   }
}
