package org.jboss.cache.api.pfer;

import org.jboss.cache.config.Configuration;
import org.testng.annotations.Test;

@Test(groups = {"functional"})
public class PFEROptimisticReplSyncTest extends PutForExternalReadTestBase
{
   public PFEROptimisticReplSyncTest()
   {
      optimistic = true;
      cacheMode = Configuration.CacheMode.REPL_SYNC;
   }
}
