package org.jboss.cache.notifications.event;

import org.jgroups.View;

/**
 * This event is passed in to any method annotated with {@link org.jboss.cache.notifications.annotation.ViewChanged}.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani</a>
 * @since 2.0.0
 */
public interface ViewChangedEvent extends Event
{
   /**
    * @return the new view associated with this view change.
    */
   View getNewView();
}
