package org.jboss.cache.commands.read;

import org.jboss.cache.Fqn;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.commands.Visitor;

/**
 * Checks whether a given node exists in current in-memory state of the cache.
 * Does not acquire any locks in doing so (result may be dirty read). Does
 * not attempt to load nodes from a cache loader (may return false if a
 * node has been evicted).
 * <p/>
 * Specifically used by the {@link org.jboss.cache.loader.ClusteredCacheLoader} to
 * implement {@link org.jboss.cache.loader.CacheLoader#exists(org.jboss.cache.Fqn)}
 * <p/>
 * This is the equivalent of the old MethodCallDefinitions.existsMethod method call from 2.1.x.
 * <p/>
 *
 * @author Mircea.Markus@jboss.com
 * @since 2.2.0
 */
public class ExistsCommand extends AbstractDataCommand
{
   public static final int METHOD_ID = 16;

   public ExistsCommand(Fqn fqn)
   {
      this.fqn = fqn;
   }

   public ExistsCommand()
   {
   }

   /**
    * Checks whether a node represented by a given Fqn exists.
    * Deleted and invalid nodes are not considered.
    *
    * @return true if the node exists, false otherwise.
    */
   public Object perform(InvocationContext ctx)
   {
      return dataContainer.exists(fqn);
   }

   public Object acceptVisitor(InvocationContext ctx, Visitor visitor) throws Throwable
   {
      return visitor.visitExistsNodeCommand(ctx, this);
   }

   public int getCommandId()
   {
      return METHOD_ID;
   }
}
