package org.jboss.cache.loader;

import org.testng.annotations.Test;

@Test(groups = {"functional"})
public class ClusteredCacheLoaderRegionBasedTest extends ClusteredCacheLoaderTest
{
   public ClusteredCacheLoaderRegionBasedTest()
   {
      useRegionBasedMarshalling = true;
   }
}
