package org.jboss.cache.lock;

/**
 * An enumeration to define different types of locks.
 *
 * @author Manik Surtani (<a href="mailto:manik@jboss.org">manik@jboss.org</a>)
 * @since 2.2.0
 */
public enum LockType
{
   NONE, READ, WRITE
}
