package org.jboss.cache.optimistic;

import org.testng.annotations.Test;

/**
 * @author <a href="mailto:manik@jboss.org">Manik Surtani</a>
 * @since 2.0.0
 */
@Test(groups = "functional")
public class LockParentVersionTest extends ParentVersionTest
{
   public LockParentVersionTest()
   {
      lockParentForChildInsertRemove = true;
   }
}
