package org.jboss.cache.commands;

import org.jboss.cache.Fqn;

/**
 * Commands of this type manipulate data in the cache.
 *
 * @author Mircea.Markus@jboss.com
 * @author Manik Surtani (<a href="mailto:manik@jboss.org">manik@jboss.org</a>)
 * @since 2.2.0
 */
public interface DataCommand extends VisitableCommand
{
   /**
    * Returns the Fqn of the node on which this command operates.
    *
    * @return an Fqn
    */
   Fqn getFqn();
}
