package org.jboss.cache.loader;

import org.jboss.cache.config.CacheLoaderConfig;

/**
 * Cache loader configuration for JDBCCacheLoaderOld implementation.
 *
 * @author <a href="mailto:galder.zamarreno@jboss.com">Galder Zamarreno</a>
 * @deprecated please use JDBCCacheLoaderConfig
 */
@Deprecated
@SuppressWarnings("deprecation")
public class JDBCCacheLoaderOldConfig extends AdjListJDBCCacheLoaderConfig
{
   /**
    * The serialVersionUID
    */
   private static final long serialVersionUID = -2911675876306188529L;

   public JDBCCacheLoaderOldConfig()
   {
      setClassName(JDBCCacheLoaderOld.class.getName());
   }

   JDBCCacheLoaderOldConfig(CacheLoaderConfig.IndividualCacheLoaderConfig base)
   {
      super(base);
      setClassName(JDBCCacheLoaderOld.class.getName());
   }
}