package org.jboss.cache.eviction.minttl;

import org.testng.annotations.Test;
import org.jboss.cache.eviction.EvictionPolicyConfigBase;
import org.jboss.cache.eviction.LRUConfiguration;
import org.jboss.cache.eviction.LFUConfiguration;

/**
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani</a>
 * @since 2.1.0
 */
@Test(groups = {"functional"})
public class LFUMinTTLTest extends MinTTLTestBase
{
   @Override
   protected EvictionPolicyConfigBase getEvictionPolicyConfig()
   {
      LFUConfiguration cfg = new LFUConfiguration();
      return cfg;
   }
}
