package org.jboss.cache.api;

import org.testng.annotations.Test;


/**
 * Optimistically locked version of {@link org.jboss.cache.api.CacheAPITest}
 */
@Test(groups = "functional")
public class CacheAPIOptimisticTest extends CacheAPITest
{
   public CacheAPIOptimisticTest()
   {
      optimistic = true;
   }
}
