/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.options.cachemodelocal;

import org.jboss.cache.config.Configuration;
import org.testng.annotations.Test;

@Test(groups = {"functional", "jgroups"})
public class SyncInvalidationOptLocksTest extends CacheModeLocalTestBase
{
    public SyncInvalidationOptLocksTest()
    {
        cacheMode = Configuration.CacheMode.INVALIDATION_SYNC;
        nodeLockingScheme = "OPTIMISTIC";
        isInvalidation = true;
    }

    protected void delay()
    {
    }
}
