package org.jboss.cache;

/**
 * Thrown when a member is suspected during remote method invocation
 *
 * @author Bela Ban
 * @version $Id: SuspectException.java 5498 2008-04-03 12:30:40Z manik.surtani@jboss.com $
 */
public class SuspectException extends CacheException
{

   private static final long serialVersionUID = -2965599037371850141L;

   public SuspectException()
   {
      super();
   }

   public SuspectException(String msg)
   {
      super(msg);
   }

   public SuspectException(String msg, Throwable cause)
   {
      super(msg, cause);
   }
}
