// $Id: ReplicationException.java 6120 2008-06-30 11:58:59Z manik.surtani@jboss.com $

/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */

package org.jboss.cache;

/**
 * Thrown when a replication problem occurred
 */
public class ReplicationException extends CacheException
{

   private static final long serialVersionUID = 33172388691879866L;

   public ReplicationException()
   {
      super();
   }

   public ReplicationException(Throwable cause)
   {
      super(cause);
   }

   public ReplicationException(String msg)
   {
      super(msg);
   }

   public ReplicationException(String msg, Throwable cause)
   {
      super(msg, cause);
   }
}
