package org.jboss.cache.notifications.event;

import org.jboss.cache.Fqn;

/**
 * Transactional events that additionally expose an Fqn as such events pertain to a specific node.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani</a>
 * @since 2.0.0
 */
public interface NodeEvent extends TransactionalEvent
{
   /**
    * @return the Fqn pointing to the node that is affected.
    */
   Fqn getFqn();
}
