// $Id: NodeNotExistsException.java 4252 2007-08-14 22:38:44Z jason.greene@jboss.com $

/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */

package org.jboss.cache;




/**
 * Thrown when an operation is attempted on a non-existing node in the cache
 *
 * @author <a href="mailto:bela@jboss.com">Bela Ban</a>.
 * @version $Id: NodeNotExistsException.java 4252 2007-08-14 22:38:44Z jason.greene@jboss.com $
 */

public class NodeNotExistsException extends CacheException {

   private static final long serialVersionUID = 779376138690777440L;

   public NodeNotExistsException() {
      super();
   }


   public NodeNotExistsException(String msg) {
      super(msg);
   }


   public NodeNotExistsException(String msg, Throwable cause) {
      super(msg, cause);
   }



}
