/**
 *
 * @author Bela Ban Nov 25, 2003
 * @version $Id: IsolationLevel.java 3756 2007-04-17 07:47:54Z msurtani $
 */
package org.jboss.cache.lock;

/**
 * Various transaction isolation levels as an enumerated class.
 *
 * @see <a href="http://en.wikipedia.org/wiki/Isolation_%28computer_science%29">Isolation levels</a>
 */
public enum IsolationLevel
{
   /**
    * No isolation.
    */
   NONE,
   SERIALIZABLE,
   REPEATABLE_READ,
   READ_COMMITTED,
   READ_UNCOMMITTED
}