/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.api.pfer;

import org.jboss.cache.config.Configuration;
import org.testng.annotations.Test;

/**
 * Test putForExternalRead with optimistic locking and INVALIDATION_ASYNC.
 * 
 * @author Brian Stansberry
 * @version $Revision$
 */
@Test (groups = {"functional"})
public class PFEROptimisticInvalidationAsyncTest extends PutForExternalReadTestBase
{
   public PFEROptimisticInvalidationAsyncTest()
   {
      optimistic = true;
      cacheMode = Configuration.CacheMode.INVALIDATION_ASYNC;
   }
}
