/*
 * JBoss, the OpenSource J2EE webOS
 * 
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.statetransfer;

import org.jboss.cache.Node;

import java.io.ObjectInputStream;

public interface StateTransferIntegrator
{

   void integrateState(ObjectInputStream ois, Node target) throws Exception;

}