package org.jboss.cache.marshall;

/**
 * MarshRegion name not found.
 *
 * @author Ben Wang 8-2005
 */
public class RegionNotFoundException extends Exception
{
 
   private static final long serialVersionUID = 7321547249627414140L;

   public RegionNotFoundException()
   {
      super();
   }

   public RegionNotFoundException(String msg)
   {
      super(msg);
   }

   public RegionNotFoundException(String msg, Throwable cause)
   {
      super(msg, cause);
   }
}
