package org.jboss.cache.loader.tcp;

/**
 * Statics that used to exist in the DelegatingCacheLoader class.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani</a>
 * @since 2.0.0
 */
public interface TcpCacheOperations
{
   int GET_CHILDREN_NAMES = 1;
   int GET_KEY = 2;
   int GET = 3;
   int EXISTS = 4;
   int PUT_KEY_VAL = 5;
   int PUT = 6;
   int REMOVE_KEY = 7;
   int REMOVE = 8;
   int REMOVE_DATA = 9;
   int LOAD_ENTIRE_STATE = 10;
   int STORE_ENTIRE_STATE = 11;
   int PUT_LIST = 12;
}
