package org.jboss.cache.eviction;

/**
 * MarshRegion name conflicts with pre-existing regions.
 *
 * @author Ben Wang 2-2004
 */
public class RegionNameConflictException extends Exception
{

   private static final long serialVersionUID = -6796150029431162837L;

   public RegionNameConflictException()
   {
      super();
   }

   public RegionNameConflictException(String msg)
   {
      super(msg);
   }

   public RegionNameConflictException(String msg, Throwable cause)
   {
      super(msg, cause);
   }
}
