package org.jboss.cache.notifications;

import org.jboss.cache.CacheException;

/**
 * Thrown when an incorrectly annotated class is added as a cache listener using the {@link org.jboss.cache.Cache#addCacheListener(Object)} API.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani</a>
 * @since 2.0.0
 */
public class IncorrectCacheListenerException extends CacheException
{

   private static final long serialVersionUID = 3847404572671886703L;

   public IncorrectCacheListenerException(String s)
   {
      super(s);
   }
}
