package org.jboss.cache.transaction;


import javax.transaction.TransactionManager;
import javax.transaction.UserTransaction;


/**
 * Returns an instance of {@link DummyTransactionManager}.
 *
 * @author Bela Ban Sept 5 2003
 * @version $Id: DummyTransactionManagerLookup.java 4526 2007-10-02 16:22:14Z jason.greene@jboss.com $
 */
public class DummyTransactionManagerLookup implements TransactionManagerLookup {

   public TransactionManager getTransactionManager() throws Exception {
      return DummyTransactionManager.getInstance();
   }

   public UserTransaction getUserTransaction()
   {
      return DummyTransactionManager.getUserTransaction();
   }

   public void cleanup()
   {
      DummyTransactionManager.destroy();
   }
}
