package org.jboss.cache.marshall;

import org.testng.annotations.Test;

/**
 * @author Manik Surtani (<a href="mailto:manik@jboss.org">manik@jboss.org</a>)
 * @since 2.1.0
 */
@Test(groups = {"functional", "jgroups"})
public class ReturnValueMarshallingMarshalledValuesTest extends ReturnValueMarshallingTest
{
   public ReturnValueMarshallingMarshalledValuesTest()
   {
      useMarshalledValues = true;
   }
}
