package org.jboss.cache.lock;

import java.util.concurrent.locks.ReentrantReadWriteLock;

/**
 * @author Bela Ban
 * @version $Id: SimpleReadWriteLock.java 3152 2006-12-11 21:14:35Z genman $
 */
public class SimpleReadWriteLock extends ReentrantReadWriteLock {

   /**
    * It's unclear why this class should be serializable.
    */
   private static final long serialVersionUID = 6317491956891614462L;

}
