/*
 * JBoss, the OpenSource J2EE webOS
 * 
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.statetransfer;

import org.jboss.cache.Node;

import java.io.ObjectOutputStream;

public interface StateTransferGenerator
{

   void generateState(ObjectOutputStream stream, Node rootNode, boolean generateTransient,
                      boolean generatePersistent, boolean suppressErrors) throws Throwable;

}