package org.jboss.cache.notifications;

import org.testng.annotations.Test;

/**
 * optimistic counterpart of {@link org.jboss.cache.notifications.RemoteCacheListenerTest}
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani</a>
 * @since 2.0.0
 */
@Test(groups = {"functional"})
public class RemoteCacheListenerOptimisticTest extends RemoteCacheListenerTest
{
   public RemoteCacheListenerOptimisticTest()
   {
      optLocking = true;
   }
}