package org.jboss.cache.transaction;


import javax.transaction.TransactionManager;


/**
 * Returns an instance of {@link BatchModeTransactionManager}.
 *
 * @author Bela Ban Sept 5 2003
 * @version $Id: BatchModeTransactionManagerLookup.java 3550 2007-02-07 22:13:24Z genman $
 */
public class BatchModeTransactionManagerLookup implements TransactionManagerLookup {

   public TransactionManager getTransactionManager() throws Exception {
      return BatchModeTransactionManager.getInstance();
   }
}
