/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.api;

import org.testng.annotations.Test;


public class NodeMoveOptimisticTest extends NodeMoveAPITest
{
   public NodeMoveOptimisticTest()
   {
      optimistic = true;
   }

   public void testLocks()
   {
      // no op
   }

   public void testLocksDeepMove()
   {
      // no op
   }

   @Override
   @Test(groups = {"functional"})
   public void testWithCacheloaders() throws Exception
   {
      super.testWithCacheloaders();    //To change body of overridden methods use File | Settings | File Templates.
   }

   @Override
   @Test(groups = {"functional"})
   public void testWithPassivation() throws Exception
   {
      super.testWithPassivation();    //To change body of overridden methods use File | Settings | File Templates.
   }
}
