package org.jboss.cache.notifications.event;

import java.util.Map;

/**
 * This event is passed in to any method annotated with {@link org.jboss.cache.notifications.annotation.NodeActivated}.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani</a>
 * @since 2.0.0
 */
public interface NodeActivatedEvent extends NodeEvent
{
   /**
    * @return an unmodifiable {@link java.util.Map} of data being activated.  Empty map when {@link #isPre()} returns <tt>true</tt>.
    */
   Map getData();
}
