/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache;

import org.jboss.cache.jmx.CacheJmxWrapperMBean;


/**
 * MBean interface.
 *
 * @author <a href="mailto:bela@jboss.org">Bela Ban</a> March 27 2003
 */
public interface TreeCacheViewMBean
{
   void create() throws Exception;

   void start() throws Exception;

   void stop() ;

   void destroy() ;
   
   Cache getCache();
   
   void setCache(Cache cache);

   CacheJmxWrapperMBean getCacheService();

   void setCacheService(CacheJmxWrapperMBean cache_service);
}
