/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.api.pfer;

import org.jboss.cache.config.Configuration;

/**
 * Test putForExternalRead with pessimistic locking and INVALIDATION_ASYNC.
 * 
 * @author Brian Stansberry
 * @version $Revision$
 */
public class PFERPessimisticInvalidationSyncTest extends PutForExternalReadTestBase
{
   public PFERPessimisticInvalidationSyncTest()
   {
      optimistic = false;
      cacheMode = Configuration.CacheMode.INVALIDATION_SYNC;
   }  
   
}
