package org.jboss.cache.eviction;

/**
 * @author Ben Wang, Feb 11, 2004
 */
public class EvictionException extends Exception
{
 
   private static final long serialVersionUID = 4006783737166646935L;

   public EvictionException()
   {
      super();
   }

   public EvictionException(String msg)
   {
      super(msg);
   }

   public EvictionException(String msg, Throwable cause)
   {
      super(msg, cause);
   }

}
