package org.jboss.cache.marshall;

public class NodeDataMarker extends NodeData
{

   private static final long serialVersionUID = 4851793846346021014L;

   @Override
   public boolean isMarker()
   {
      return true;
   }

   @Override
   public String toString()
   {
      return "NodeDataMarker";
   }
}
