package org.jboss.cache;

/**
 * Exception to represent a region being empty when state was expected in that region.
 *
 * @author Manik Surtani (<a href="mailto:manik@jboss.org">manik@jboss.org</a>)
 * @since 2.2.0
 */
public class RegionEmptyException extends CacheException
{
   public RegionEmptyException()
   {
   }

   public RegionEmptyException(Throwable cause)
   {
      super(cause);
   }

   public RegionEmptyException(String msg)
   {
      super(msg);
   }

   public RegionEmptyException(String msg, Throwable cause)
   {
      super(msg, cause);
   }
}
