package org.jboss.cache.notifications.event;

import org.jboss.cache.buddyreplication.BuddyGroup;

/**
 * This event is passed in to any method annotated with {@link org.jboss.cache.notifications.annotation.BuddyGroupChanged}.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani</a>
 * @since 2.1.0
 */
public interface BuddyGroupChangedEvent extends Event
{
   /**
    * @return the new buddy group
    * @since 2.1.0
    */
   BuddyGroup getBuddyGroup();
}