package org.jboss.cache.notifications.event;

import org.jboss.cache.Fqn;

/**
 * This event is passed in to any method annotated with {@link org.jboss.cache.notifications.annotation.NodeMoved}.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani</a>
 * @since 2.0.0
 */
public interface NodeMovedEvent extends NodeEvent
{
   /**
    * @return the new, resultant Fqn after the move
    */
   Fqn getTargetFqn();
}
